/*==============================================================================
 * PinIO.h -- classes for digital input/output pin state management.
 *============================================================================*/

#ifndef PINIO_H
#define PINIO_H

enum PinState { LOW, HIGH };
enum PinEdge { NONE, FALLING, RISING, IGNORE };

// convenience function to swap rising/falling edge value or return unchanged
PinEdge edgeSwap(PinEdge edge) {
  if (edge == PinEdge::NONE) return edge;
  return edge == PinEdge::RISING ? PinEdge::FALLING : PinEdge::RISING;
}

/*------------------------------------------------------------------------------
 * PinBase class -- class to manage output pin states.
 *----------------------------------------------------------------------------*/
class PinBase {
public:
  // PinBase(){};
  PinBase(PinState state = PinState::LOW, PinEdge edge = PinEdge::NONE)
      : state(state), edge(edge) {}

public:
  PinState getState() const { return state; }

  PinEdge getEdge() const { return edge; }

  PinState setState(PinState toState) {
    if (state == toState) edge = PinEdge::NONE;
    else edge = toState == PinState::HIGH ? PinEdge::RISING : PinEdge::FALLING;
    state = toState;
    return state;
  }

  PinState setState(bool toState) {
    return setState(toState ? PinState::HIGH : PinState::LOW);
  }

  inline PinState toggleState() {
    return setState(state == PinState::HIGH ? PinState::LOW : PinState::HIGH);
  }

  // convenience methods
  inline bool isHigh() const { return state == PinState::HIGH; }
  inline bool isLow() const { return !isHigh(); }
  inline bool isEdge() const { return edge != PinEdge::NONE; }
  inline bool isRising() const { return edge == PinEdge::RISING; }
  inline bool isFalling() const { return edge == PinEdge::FALLING; }
  PinState    setHigh() { return setState(PinState::HIGH); }
  PinState    setLow() { return setState(PinState::LOW); }

protected:
  PinState state;   // = PinState::LOW;
  PinEdge  edge;    //  = NONE;
};

/*------------------------------------------------------------------------------
 * PinOut class -- class to manage output pin states.
 *
 * This class saves the current state and can return the current state as a
 * digital "high" or "low" voltage.
 *----------------------------------------------------------------------------*/
class PinOut : public PinBase {
public:
  PinOut() : PinBase() {}
  PinOut(double vcc, PinState idleState) : PinBase(idleState) {
    init(vcc, idleState);
  }

protected:
  void init(double vcc, PinState idleState) {
    vHigh           = vcc;
    this->idleState = idleState;
  }

public:
  double getStateV() const { return state == PinState::HIGH ? vHigh : 0.0; }

  void setIdle() { setState(idleState); }

protected:
  double   vHigh;
  PinState idleState;
};

/*------------------------------------------------------------------------------
 * PinIn class -- class to manage input pin states.
 *
 * This class saves the current pin state and, when changed, also saves the
 * transition information (i.e., that the last state change was a rising/falling
 * edge or no change/not an edge transition).
 *
 * Note:  It would be easy to add hysteresis to this class if that would be
 * useful.
 *----------------------------------------------------------------------------*/
class PinIn : public PinBase {
public:
  PinIn() : PinBase() {}
  PinIn(double vcc, double vIn) : PinBase() { init(vcc, vIn); }

protected:
  void init(double vcc, double vIn) {
    halfVcc = vcc / 2.0;
    setState(vIn);
  }

public:
  PinState setState(double vPin) {
    // why must I manually disambiguate???  or do I???
    return PinBase::setState(
        PinState(vPin > halfVcc ? PinState::HIGH : PinState::LOW));
  }

protected:
  double halfVcc = 0.0;
};

#endif   // PINIO_H