#ifndef SERIALBUFFER_H
#define SERIALBUFFER_H

#include <cinttypes>

void msg(int lineNbr, const char *fmt, ...);   // fwd decl

/*------------------------------------------------------------------------------
 * class SerialBuffer - a single buffer for both sending & recieving data.
 * buffer is hard-coded for 32 bits but that could easily be increased to 64
 * bits.  data rotates out of the MSb and in to the LSb.  the number of bits to
 * be sent in a transaction is set in the startIO() method
 *----------------------------------------------------------------------------*/
class SerialBuffer {
public:
  // sets initial data to send and # of bits for transaction
  void startIO(uint32_t startData, uint8_t bits) {
    data      = startData;
    bitsTotal = bits;
    bitsIn    = 0;
    msbMask   = 1 << (bits - 1);
    data      = startData & ~(UINT32_MAX << bits);
    overFlow  = false;

    // msg(__LINE__, "startIO() data=%0.2x, bits=%d, msbMask=%x\n", data, bits,
    //     msbMask);
  }

  void endIO() { /* nothing to do? */
  }

  // set the buffer data -- useful if output data isn't known when starting the
  // transaction.  the new buffer output data is shifted left by the number of
  // bits already sent and preserves any data already received.
  void setData(uint32_t newData) {
    data = getData();            // clear bits not yet sent
    data |= newData << bitsIn;   // set unsent bits
  }

  // get the MSb to send -- no need to call if no output needed
  inline bool getBitOut() { return data & msbMask; }

  // rotate bit into LSb -- must call this between getBitOut() calls to advance
  // the data bits even if there's no data input
  void setBitIn(bool bit) {   // = false) {
    if (isDone()) {
      // this is an error state...
      overFlow = true;
      return;
    }
    data <<= 1;
    data |= bit ? 1 : 0;
    bitsIn++;
  }

  // get the current data received.  note that, if the transaction isn't
  // complete, only the received bits are returned.  this can be useful if
  // the first few bits received in the transaction determine what bits will
  // be sent in the remaining bits of the transaction.
  uint32_t getData() const {
    uint32_t mask = ~(UINT32_MAX << bitsIn);
    return data & mask;
  }

  inline bool    isDone() const { return bitsIn >= bitsTotal; }
  inline bool    isOverflow() const { return overFlow; }
  inline uint8_t getBitsIn() const { return bitsIn; }

protected:
  uint32_t data;
  uint32_t msbMask;
  uint8_t  bitsTotal;   // transaction size in bits
  uint8_t  bitsIn;      // # of bits input
  bool     overFlow = false;
};

#endif   // SERIALBUFFER_H