/*==============================================================================
 * SpiDAC.cpp -- Demonstration DAC SPI slave device.
 *============================================================================*/
// Note:  Compile with MS VC

#include <stdio.h>
#include <stdarg.h>
#include <thread>

#include "SerialBuffer.h"
#include "PinIO.h"

// versioning
#define PROGRAM_NAME    "SpiDAC"
#define PROGRAM_VERSION "v0.1"
#define PROGRAM_INFO    PROGRAM_NAME " " PROGRAM_VERSION

union uData {
  bool                   b;
  char                   c;
  unsigned char          uc;
  short                  s;
  unsigned short         us;
  int                    i;
  unsigned int           ui;
  float                  f;
  double                 d;
  long long int          i64;
  unsigned long long int ui64;
  char                  *str;
  unsigned char         *bytes;
};

// #undef pin names lest they collide with names in any header file(s) you might
// include.
#undef CS
#undef SCLK
#undef VCC
#undef MOSI
#undef MISO
#undef DAC_OUT

// int DllMain() must exist and return 1 for a process to load the .DLL
// See https://docs.microsoft.com/en-us/windows/win32/dlls/dllmain for more
// information.
int __stdcall DllMain(void *module, unsigned int reason, void *reserved) {
  return 1;
}

#define msleep(msecs)                                                          \
  std::this_thread::sleep_for(std::chrono::milliseconds(msecs))

void msg(int lineNbr, const char *fmt, ...) {
  msleep(30);
  fflush(stdout);
  fprintf(stdout, PROGRAM_INFO " (@%d) ", lineNbr);
  va_list args = {0};
  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
  fflush(stdout);
  msleep(30);
}

/*------------------------------------------------------------------------------
 * The SPI Mode stuff is complicated...
 *
 * SPI  | Clk Polarity
 * Mode | CPOL CPHA | Data is shifted out on              | Data is sampled on
 *   0  |   0    0  | Falling SCLK, and when CS activates | Rising SCLK
 *   1  |   0    1  | Rising SCLK                         | Falling SCLK
 *   2  |   1    0  | Rising SCLK, and when CS activates  | Falling SCLK
 *   3  |   1    1  | Falling SCLK                        | Rising SCLK
 *----------------------------------------------------------------------------*/

struct SpiModeTbl {
  PinState sclkIdle;
  PinEdge  csOutEdge;
  PinEdge  sclkOutEdge;
  PinEdge  sclkInEdge;
};

const SpiModeTbl spiModes[] = {
    // mode 0 - CPOL=0  | Falling SCLK, and when CS activates | Rising SCLK
    {
        PinState::LOW,      // sclkIdle
        PinEdge::FALLING,   // csOutEdge (CS hardcoded for active low)
        PinEdge::FALLING,   // sclkOutEdge
        PinEdge::RISING     // sclkInEdge
    },
    // mode 1 - CPOL=0 | Rising SCLK | Falling SCLK
    {
        PinState::LOW,     // sclkIdle
        PinEdge::IGNORE,   // csOutEdge
        PinEdge::RISING,   // sclkOutEdge
        PinEdge::FALLING   // sclkInEdge
    },
    // mode 2 - CPOL=1 | Rising SCLK, and when CS activates | Falling SCLK
    {
        PinState::HIGH,     // sclkIdle
        PinEdge::FALLING,   // csOutEdge (CS hardcoded for active low)
        PinEdge::RISING,    // sclkOutEdge
        PinEdge::FALLING    // sclkInEdge
    },
    // mode 3 - CPOL=1 | Falling SCLK | Rising SCLK
    {
        PinState::HIGH,     // sclkIdle
        PinEdge::IGNORE,    // csOutEdge
        PinEdge::FALLING,   // sclkOutEdge
        PinEdge::RISING     // sclkInEdge
    }};

// enum SpiState { Idle, Initialized, Active };
const unsigned int SpiFreqDef = 10000;   // default speed if attribute invalid
const unsigned int SpiModeDef = 3;   // default SPI mode if attribute invalid

struct InstData {
  PinIn        sclkPinIn;    // system clock
  PinIn        csPinIn;      // chip select signal
  PinOut       misoPinOut;   // MISO pin
  PinIn        mosiPinIn;    // MOSI pin
  SerialBuffer sBuf;         // output buffer
  SpiModeTbl   spiMode;      // SPI mode for instance
  double       dacOutV;      // DAC output pin
};

// fwd decls
// bool loadDataBuf(InstData &inst);
void loadDataBuf(InstData &inst, uData *data);
void bitReceived(InstData &inst, uData *data);
void processDataBuf(InstData &inst, uData *data);

// we use the uData struct in several places so make a #define for convenience
#define UDATA(data)                                                            \
  double  CS      = data[0].d;                                                 \
  double  SCLK    = data[1].d;                                                 \
  double  VCC     = data[2].d;                                                 \
  double  MOSI    = data[3].d;                                                 \
  int     SPIMODE = data[4].i;                                                 \
  double &MISO    = data[5].d;                                                 \
  double &DAC_OUT = data[6].d;

extern "C" __declspec(dllexport) void spidac(
    InstData **opaque, double t, uData *data) {
  UDATA(data);

  InstData *inst = *opaque;

  if (!inst) {
    // first time, VCC is 0.0V so delay until VCC is something valid...
    if (VCC == 0.0) return;

    *opaque = inst = new InstData();
    if (!inst) {   // terminate with prejudice
      msg(__LINE__, "Unable to allocate memory.  Terminating simulation.\n");
      std::terminate();
    }

    // SPI mode could be an attribute from component instance... for now,
    // hardcode
    // copy SPI mode table to instance data
    // const uint8_t spiMode = 3;
    // inst->spiMode         = spiModes[spiMode];
    if (SPIMODE < 0 || SPIMODE > 3) {
      msg(__LINE__,
          "SpiMode=%d is not valid.  Valid values are 0-3.  Using default "
          "mode=%d.\n",
          SPIMODE, SpiModeDef);
      SPIMODE = SpiModeDef;
    }
    inst->spiMode = spiModes[SPIMODE];

    // configure some pins
    inst->csPinIn    = PinIn(VCC, CS);
    inst->misoPinOut = PinOut(VCC, PinState::LOW);
    inst->mosiPinIn  = PinIn(VCC, MOSI);
    DAC_OUT          = 0.0;   // start at 0V

    // debug info
    msg(__LINE__, "SpiMode=%d.\n", SPIMODE);

    // for now, just return after initialization?
    return;
  }

  // set PinIn states from inputs
  inst->csPinIn.setState(CS);
  inst->mosiPinIn.setState(MOSI);
  inst->sclkPinIn.setState(SCLK);

  // set output pins from last PinOut states (necessary?)
  MISO = inst->misoPinOut.getStateV();

  if (inst->csPinIn.isHigh() && !inst->csPinIn.isRising()) { return; }

  // if just now enabled, set up for start of SPI I/O
  // note that SCLK must be in idle state because wasn't enabled
  if (inst->csPinIn.isFalling()) {
    // load data
    loadDataBuf(*inst, data);
  }

  // read MOSI
  if (inst->sclkPinIn.getEdge() == inst->spiMode.sclkInEdge) {
    inst->sBuf.setBitIn(inst->mosiPinIn.isHigh());
    if (inst->sBuf.isDone()) {
      // process data
      processDataBuf(*inst, data);

      // set MOSI to idle
      inst->misoPinOut.setIdle();
      MISO = inst->misoPinOut.getStateV();
    } else bitReceived(*inst, data);
    return;
  }

  // set MISO
  if (inst->csPinIn.isRising()) return;
  if (inst->sclkPinIn.getEdge() == inst->spiMode.sclkOutEdge ||
      inst->csPinIn.getEdge() == inst->spiMode.csOutEdge) {
    inst->misoPinOut.setState(inst->sBuf.getBitOut());
    MISO = inst->misoPinOut.getStateV();
    return;
  }
}

// TODO: This returns a value only if master...
// this is where you add code for data to send
void loadDataBuf(InstData &inst, uData *data) {
  UDATA(data);

  // load data here
  // nothing to load for this device but must set up buffer
  inst.sBuf.startIO(0, 8);
}

// this is called for each bit received -- if the remaining bits to be sent
// depend on bits received so far, change them here
void bitReceived(InstData &inst, uData *data) {
  UDATA(data);

  // load data here
  /* nothing to do for this device */
}

// this is where you do something with the data from a completed transaction
void processDataBuf(InstData &inst, uData *data) {
  UDATA(data);

  // set output voltage (VCC-referenced)
  uint8_t vOut8 = (uint8_t)(inst.sBuf.getData());
  double  x     = vOut8;
  DAC_OUT = inst.dacOutV = x * VCC / 0xff;
  // msg(__LINE__, "DAC received %x, DAC_OUT=%f\n", vOut8, DAC_OUT);
}

extern "C" __declspec(dllexport) void Destroy(struct InstData *inst) {
  delete inst;
}
/*==============================================================================
 * End of SpiDAC.cpp
 *============================================================================*/