/*==============================================================================
 * SpiMaster.cpp -- Generic Spi master device.
 *============================================================================*/
// Note:  Compile with MS VC

#include <stdio.h>
#include <stdarg.h>
#include <thread>

#include "SerialBuffer.h"
#include "PinIO.h"

// versioning
#define PROGRAM_NAME    "SpiMaster"
#define PROGRAM_VERSION "v0.1"
#define PROGRAM_INFO    PROGRAM_NAME " " PROGRAM_VERSION

union uData {
  bool                   b;
  char                   c;
  unsigned char          uc;
  short                  s;
  unsigned short         us;
  int                    i;
  unsigned int           ui;
  float                  f;
  double                 d;
  long long int          i64;
  unsigned long long int ui64;
  char                  *str;
  unsigned char         *bytes;
};

// #undef pin names lest they collide with names in any header file(s) you might
// include.
#undef EN
#undef SCLK
#undef MOSI
#undef MISO
#undef SS1
#undef SS2
#undef VCC

#define UDATA                                                                  \
  double  EN      = data[0].d;                                                 \
  double  MISO    = data[1].d;                                                 \
  double  VCC     = data[2].d;                                                 \
  int     SPIFREQ = data[3].i;                                                 \
  int     SPIMODE = data[4].i;                                                 \
  double &SCLK    = data[5].d;                                                 \
  double &MOSI    = data[6].d;                                                 \
  double &SS1     = data[7].d;                                                 \
  double &SS2     = data[8].d;

// int DllMain() must exist and return 1 for a process to load the .DLL
// See https://docs.microsoft.com/en-us/windows/win32/dlls/dllmain for more
// information.
int __stdcall DllMain(void *module, unsigned int reason, void *reserved) {
  return 1;
}

#define msleep(msecs)                                                          \
  std::this_thread::sleep_for(std::chrono::milliseconds(msecs))

void msg(int lineNbr, const char *fmt, ...) {
  msleep(30);
  fflush(stdout);
  fprintf(stdout, PROGRAM_INFO " (@%d) ", lineNbr);
  va_list args = {0};
  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
  fflush(stdout);
  msleep(30);
}

// /*------------------------------------------------------------------------------
//  * The SPI Mode stuff is complicated...
//  *
//  * SPI  | Clk Polarity
//  * Mode | CPOL CPHA | Data is shifted out on              | Data is sampled
//  on
//  *   0  |   0    0  | Falling SCLK, and when CS activates | Rising SCLK
//  *   1  |   0    1  | Rising SCLK                         | Falling SCLK
//  *   2  |   1    0  | Rising SCLK, and when CS activates  | Falling SCLK
//  *   3  |   1    1  | Falling SCLK                        | Rising SCLK
//  *----------------------------------------------------------------------------*/

struct SpiModeTbl {
  PinState sclkIdle;
  PinEdge  csOutEdge;
  PinEdge  sclkOutEdge;
  PinEdge  sclkInEdge;
};

const SpiModeTbl spiModes[] = {
    // mode 0 - CPOL=0  | Falling SCLK, and when CS activates | Rising SCLK
    {
        PinState::LOW,      // sclkIdle
        PinEdge::FALLING,   // csOutEdge (CS hardcoded for active low)
        PinEdge::FALLING,   // sclkOutEdge
        PinEdge::RISING     // sclkInEdge
    },
    // mode 1 - CPOL=0 | Rising SCLK | Falling SCLK
    {
        PinState::LOW,     // sclkIdle
        PinEdge::IGNORE,   // csOutEdge
        PinEdge::RISING,   // sclkOutEdge
        PinEdge::FALLING   // sclkInEdge
    },
    // mode 2 - CPOL=1 | Rising SCLK, and when CS activates | Falling SCLK
    {
        PinState::HIGH,     // sclkIdle
        PinEdge::FALLING,   // csOutEdge (CS hardcoded for active low)
        PinEdge::RISING,    // sclkOutEdge
        PinEdge::FALLING    // sclkInEdge
    },
    // mode 3 - CPOL=1 | Falling SCLK | Rising SCLK
    {
        PinState::HIGH,     // sclkIdle
        PinEdge::IGNORE,    // csOutEdge
        PinEdge::FALLING,   // sclkOutEdge
        PinEdge::RISING     // sclkInEdge
    }};

// enum SpiState { Idle, Initialized, Active };

const unsigned int SpiFreqDef = 10000;   // default speed if attribute invalid
const unsigned int SpiModeDef = 3;   // default SPI mode if attribute invalid

struct InstData {
  // int testCtr = 0;
  // PinIn        clkPinIn;     // system clock
  PinIn        enPinIn;      // enable signal
  PinIn        misoPinIn;    // MISO pin
  PinOut       mosiPinOut;   // MOSI pin
  PinOut       sclkPinOut;   // SPI clock
  PinOut       ss1PinOut;    // slave select
  PinOut       ss2PinOut;    // slave select
  SerialBuffer sBuf;         // output buffer

  SpiModeTbl spiMode;   // SPI mode for instance

  bool adcRead = true;   // true=reading ADC, false = writing DAC

  // for managing SCLK...
  // int sclkHz;   // internal SCLK frequency (approx)
  // double   sclkCycleT;          // full SCLK cycle seconds
  double sclkHalfCycleT;   // half SCLK cycle seconds
  // double   sclkPrePostDelayT;   // deley seconds
  double sclkNextToggleT;
};

// uint8_t testData8[] = {0x88, 0x11, 0x55, 0xaa, 0xde, 0xad, 0xbe, 0xef, 0x00};

const double eternity = 1.7e308;   // end of the 'verse

// fwd decls
bool loadDataBuf(InstData &inst);
void processDataBuf(InstData &inst);

extern "C" __declspec(dllexport) void spimaster(
    InstData **opaque, double t, uData *data) {
  UDATA;

  InstData *inst = *opaque;

  if (!inst) {
    // first time, VCC is 0.0V so delay until VCC is something valid...
    if (VCC == 0.0) return;

    *opaque = inst = new InstData();
    if (!inst) {   // terminate with prejudice
      msg(__LINE__, "Unable to allocate memory.  Terminating simulation.\n");
      std::terminate();
    }

    // set up some timing information -- SCLK frequency could be from attribute
    // but hardcode for now (master only)
    // const uint32_t spiFreq = 100000;
    // uint32_t spiFreq = spiFreq

    // inst->sclkHz            = spiFreq;   // internal SCLK frequency (approx)
    // inst->sclkCycleT        = 1.0 / spiFreq;         // seconds
    // inst->sclkHalfCycleT    = 1.0 / spiFreq / 2.0;   // seconds
    // inst->sclkPrePostDelayT = 1.0 / spiFreq / 4.0;
    // inst->sclkNextToggleT   = eternity;

    // inst->sclkHz          = SpiFreq;   // internal SCLK frequency (approx)
    if (SPIFREQ < 1) {
      msg(__LINE__, "SpiFreq=%d is not valid.  Using default frequency=%dHz.\n",
          SPIFREQ, SpiFreqDef);
      SPIFREQ = SpiFreqDef;
    }

    inst->sclkHalfCycleT  = (1.0 / SPIFREQ) / 2.0;   // seconds per half cycle
    inst->sclkNextToggleT = eternity;

    // SPI mode could be an attribute from component instance... for now,
    // hardcode
    // copy SPI mode table to instance data
    // const uint8_t spiMode = 3;
    // inst->spiMode         = spiModes[spiMode];
    if (SPIMODE < 0 || SPIMODE > 3) {
      msg(__LINE__,
          "SpiMode=%d is not valid.  Valid values are 0-3.  Using default "
          "mode=%d.\n",
          SPIMODE, SpiModeDef);
      SPIMODE = SpiModeDef;
    }
    inst->spiMode = spiModes[SPIMODE];

    inst->enPinIn = PinIn(VCC, EN);
    // inst->clkPinIn  = PinIn(VCC, CLK);
    inst->misoPinIn = PinIn(VCC, MISO);

    // initialize PinOut instances
    inst->sclkPinOut = PinOut(VCC, inst->spiMode.sclkIdle);
    inst->mosiPinOut = PinOut(VCC, PinState::LOW);
    inst->ss1PinOut  = PinOut(VCC, PinState::HIGH);
    inst->ss2PinOut  = PinOut(VCC, PinState::HIGH);

    SCLK = inst->sclkPinOut.getStateV();   // from SPI mode
    MOSI = inst->mosiPinOut.getStateV();   // doesn't really matter
    SS1  = inst->ss1PinOut.getStateV();    // active low
    SS2  = inst->ss2PinOut.getStateV();    // active low

    // debug info
    msg(__LINE__, "SpiFreq=%dHz, SpiMode=%d.\n", SPIFREQ, SPIMODE);
  }

  /*------------------------------------------------------------------------------
   * We're going to need some configuration options...
   *----------------------------------------------------------------------------*/

  // set PinIn states from inputs
  // inst->clkPinIn.setState(CLK);
  inst->enPinIn.setState(EN);
  inst->misoPinIn.setState(MISO);

  // set output pins from PinOut states (necessary?)
  SCLK = inst->sclkPinOut.getStateV();
  MOSI = inst->mosiPinOut.getStateV();
  SS1  = inst->ss1PinOut.getStateV();
  SS2  = inst->ss2PinOut.getStateV();

  if (inst->enPinIn.isRising()) {
    // interrupt any processing that might be under way & reset outputs to idle
    // TODO
    // inst->sclkNextToggleT = eternity;

    // we're really done -- set stuff to idle
    // stop SCLK
    inst->sclkNextToggleT = eternity;
    inst->sclkPinOut.setIdle();
    SCLK = inst->sclkPinOut.getStateV();

    // disable ss
    inst->ss1PinOut.setHigh();
    SS1 = inst->ss1PinOut.getStateV();
    inst->ss2PinOut.setHigh();
    SS2 = inst->ss2PinOut.getStateV();

    // set MOSI to idle
    inst->mosiPinOut.setIdle();
    MOSI = inst->mosiPinOut.getStateV();

    inst->sBuf.endIO();

    // msg(__LINE__, "Got here 3\n");
  }

  if (inst->enPinIn.isHigh()) { return; }

  // if just now enabled, set up for start of SPI I/O
  // note that SCLK must be in idle state because wasn't enabled
  if (inst->enPinIn.isFalling()) {
    // load data
    bool haveData = loadDataBuf(*inst);
    if (!haveData) return;

    // if (haveData) {
    // enable slave
    if (inst->adcRead) {
      inst->ss1PinOut.setLow();
      SS1 = inst->ss1PinOut.getStateV();
      inst->ss2PinOut.setHigh();
      SS2 = inst->ss2PinOut.getStateV();
    } else {
      inst->ss1PinOut.setHigh();
      SS1 = inst->ss1PinOut.getStateV();
      inst->ss2PinOut.setLow();
      SS2 = inst->ss2PinOut.getStateV();
    }

    // need to have a common SS here?!?
    // need to have a common SS here?!?
    // need to have a common SS here?!?
    // need to have a common SS here?!?
    inst->mosiPinOut.setState(inst->sBuf.getBitOut());
    MOSI = inst->mosiPinOut.getStateV();

    // next sim time to toggle SCLK
    inst->sclkNextToggleT = t + inst->sclkHalfCycleT;
    // return;
    // }
  }

  // toggle SCLK now?
  if (inst->sclkNextToggleT > t) return;

  inst->sclkPinOut.toggleState();
  SCLK = inst->sclkPinOut.getStateV();
  // next sim time to toggle SCLK
  inst->sclkNextToggleT = t + inst->sclkHalfCycleT;

  // set MOSI
  if (inst->sclkPinOut.getEdge() == inst->spiMode.sclkOutEdge ||
      inst->ss1PinOut.getEdge() == inst->spiMode.csOutEdge) {

    if (inst->sBuf.isDone()) {
      // we're really done -- set stuff to idle
      // stop SCLK
      inst->sclkNextToggleT = eternity;
      inst->sclkPinOut.setIdle();
      SCLK = inst->sclkPinOut.getStateV();

      // disable ss
      inst->ss1PinOut.setHigh();
      SS1 = inst->ss1PinOut.getStateV();
      inst->ss2PinOut.setHigh();
      SS2 = inst->ss2PinOut.getStateV();

      // set MOSI to idle
      inst->mosiPinOut.setIdle();
      MOSI = inst->mosiPinOut.getStateV();

      // msg(__LINE__, "Got here 3\n");

      return;
    }
    inst->mosiPinOut.setState(inst->sBuf.getBitOut());
    MOSI = inst->mosiPinOut.getStateV();
  }

  // read MISO
  if (inst->sclkPinOut.getEdge() == inst->spiMode.sclkInEdge) {
    inst->sBuf.setBitIn(inst->misoPinIn.isHigh());
    if (inst->sBuf.isDone()) {
      // process data
      processDataBuf(*inst);

      // swap read ADC / write DAC
      inst->adcRead = !inst->adcRead;

      // // stop SCLK, disable SS
      // inst->ss1PinOut.setHigh();
      // SS1 = inst->ss1PinOut.getStateV();
      // inst->ss2PinOut.setHigh();
      // SS2 = inst->ss2PinOut.getStateV();

      // inst->sclkNextToggleT = eternity;
      // inst->sclkPinOut.setIdle();
      // SCLK = inst->sclkPinOut.getStateV();

      // // set MOSI to idle
      // inst->mosiPinOut.setIdle();
      // MOSI = inst->mosiPinOut.getStateV();
    }
  }
}

extern "C" __declspec(dllexport) double MaxExtStepSize(InstData *inst) {
  // return 1e308;   // implement a good choice of max timestep size that
  // depends
  //                 // on struct sSPIMASTER

  return inst->sclkHalfCycleT;
}

extern "C" __declspec(dllexport) void Trunc(InstData *inst, double t,
    uData  *data,
    double *timestep) {   // limit the timestep to a tolerance if the circuit
  // causes a change in struct sSPIMASTER

  UDATA;

  double ttol = 1e-9;
  // if (*timestep > ttol) {
  //   double &SCLK = data[4].d;   // output
  //   double &MOSI = data[5].d;   // output
  //   double &SS1  = data[6].d;   // output
  //   double &SS2  = data[7].d;   // output

  //   // Save output vector
  //   const double _SCLK = SCLK;
  //   const double _MOSI = MOSI;
  //   const double _SS1  = SS1;
  //   const double _SS2  = SS2;

  //   // copy instance data, make pointer to that instance, and pass the **copy
  //   InstData  tmp   = *inst;
  //   InstData *_inst = &tmp;
  //   spimaster(&_inst, t, data);

  //   // if(tmp != *inst) // implement a meaningful way to detect if the state
  //   has
  //   // changed
  //   //    *timestep = ttol;

  //   // Restore output vector
  //   SCLK = _SCLK;
  //   MOSI = _MOSI;
  //   SS1  = _SS1;
  //   SS2  = _SS2;
  // }

  if (t < inst->sclkNextToggleT) ttol = inst->sclkNextToggleT - t;
  *timestep = ttol;
}

// TODO: This returns a value only if master...
// this is where you add code for data to send; return false if no data (master
// only)
bool loadDataBuf(InstData &inst) {
  // load the buffer (TODO: move to an individual method)
  // if (inst.testCtr < sizeof(testData8) / sizeof(testData8[0])) {
  //   inst.sBuf.startIO(testData8[inst.testCtr], 8);
  //   inst.testCtr++;
  //   return true;
  // }
  // return false;

  // if we're reading the ADC, just send zeros
  uint8_t bitsToSend = inst.adcRead ? 0 : (uint8_t)inst.sBuf.getData();
  inst.sBuf.startIO(bitsToSend, 8);
  return true;
}

// this is where you do something with the data from a completed transaction
void processDataBuf(InstData &inst) {
  // msg(__LINE__, "Received data=%0.2x\n", inst.sBuf.getData());
  inst.sBuf.endIO();
}

extern "C" __declspec(dllexport) void Destroy(struct InstData *inst) {
  //   free(inst);
  delete inst;
}

/*==============================================================================
 * End of SpiMaster.cpp
 *============================================================================*/